/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.cache;

import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.perseus.cache.lib.BasicCacheEntryFactory;
import org.objectweb.perseus.cache.lib.BasicCacheManager;
import org.objectweb.perseus.cache.replacement.lib.AbstractReplacementManager;
import org.objectweb.perseus.cache.replacement.lib.FIFOReplacementManager;
import org.objectweb.perseus.cache.replacement.lib.LRUReplacementManager;
import org.objectweb.perseus.cache.replacement.lib.MRUReplacementManager;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.LoggerFactory;

/**
 * This class is a factory of CacheManager, but it does not used the Julia
 * for instanciate the component. All is done manualy, in order to avoid the 
 * use of a Julia runtime (Julia is an implementation of a fractal runtime. 
 * 
 * @author S.Chassande-Barrioz
 */
public class CacheManagerFactory {

	public final static String LRU_POLICY = "LRU";
	public final static String MRU_POLICY = "MRU";
	public final static String FIFO_POLICY = "FIFO";
	public final static String CM_LOGGER_NAME ="org.objectweb.perseus.cache.cache-manager";
	public final static String RM_LOGGER_NAME ="org.objectweb.perseus.cache.replacement-manager";


	public static CacheManager newCacheManager(int size, String policy) throws Exception {
		LoggerFactory lf = Monolog.initialize();

		//Instanciate the cache primitive component
		BasicCacheManager cm = new BasicCacheManager();
		cm.bindFc("logger", lf.getLogger(CM_LOGGER_NAME));
		cm.bindFc("monolog-factory", lf);
		cm.setMaxObjects(size);
		
		//Instanciate the Cache entry factory
		BasicCacheEntryFactory cef = new BasicCacheEntryFactory();
		
		//Instanciate the replacement manager
		AbstractReplacementManager rm = getRM(policy);
		rm.bindFc("logger", lf.getLogger(RM_LOGGER_NAME));
		rm.bindFc("monolog-factory", lf);

		//Binds primitive components together
		cm.bindFc(BasicCacheManager.CACHE_ENTRY_FACTORY_BINDING, cef);
		cm.bindFc(BasicCacheManager.REPLACEMENT_MANAGER_BINDING, rm);
		rm.bindFc(AbstractReplacementManager.UNBIND_MANAGER_BINDING, cm);
		
		//start the components
		cm.startFc();
		return cm;
	}
	
	private static AbstractReplacementManager getRM(String policy) {
		if (LRU_POLICY.equals(policy)) {
			return new LRUReplacementManager();
		} else if (MRU_POLICY.equals(policy)) {
			return new MRUReplacementManager();
		} else if (FIFO_POLICY.equals(policy)) {
			return new FIFOReplacementManager();
		} else {
			return new LRUReplacementManager();
		}
	}

}
