/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.cache.api;

import org.objectweb.fractal.api.control.AttributeController;

import java.util.Collection;

/**
 * This interface is management-related. Application code should
 * not use this interface. Further, this interface is intended to be
 * used by service tuning through Ping provided tools.
 *
 * This is the first draft for the this interface and it is likely to
 * evolve.
 *
 * @author Luciano Garcia-Banuelos (Luciano.Garcia@imag.fr)
 */
public interface CacheAttributeController extends AttributeController {

    public final static int NO_LIMIT = -1;

    /**
     * This method allows to set the maximum number of objects to be
     * held by the cache manager. However, the actual capacity of
     * the cache depends on the Java Virtual Machine memory limits
     * (and eventually by the amount of disk allocated).
     * That is, if the application allocates a lot of large objects,
     * and the memory limit is reached, the cache would not be able
     * to hold more objects.
     *
     * The cache replacement task is launched when the maximum number
     * of objects limit is nearly to be reached.
     *
     * @param size The maximum number of objects to be held in cache.
     *             (It must be a positive number).
     *
     * @exception IllegalArgumentException if the size is invalid (e.g. negative).
     * @exception CacheException if it is not possible to reduce the cache size
	 * because all entries are fixed.
     */
    void setMaxObjects(int size) throws IllegalArgumentException, CacheException;

    /**
	 * This method retrieves the maximum number of objects to be held by the
	 * cache.
	 */
	int getMaxObjects();

    /**
     *  This  method  allows  to  set  the  maximum  main-memory  size
     *  allocated to  the cache.  The size  is limited  by  the actual
     * amount of memory allocated to the Java Virtual Machine.
     *
     * @param size The maximum memory size.
     *             (It must be a positive number).
     *
     * @exception IllegalArgumentException if the size is invalid (e.g. negative).
     */
    void setMemorySize(int size) throws IllegalArgumentException;

	/**
     *  This  method  returns the  maximum  main-memory  size
     *  allocated to  the cache.
	 */
    int getMemorySize();

	/**
	 * Assignes the number of element which can be removed when the cache is
	 * full. This value can be an absolute value( ex: "124") or a percent of the
	 * maximal cache size ("8%").
	 */ 
	void setAutoCleanSize(String size);

	String getAutoCleanSize();

	/**
	 * Assignes the cache size value since the cache try to decrease the number
	 * of entries. This value can be an absolute value( ex: "124") or a percent
	 * of the maximal cache size ("8%").
	 */
	void setAutoCleanThreshold(String size);

	String getAutoCleanThreshold();

	/**
	 * @return an positive integer value. It is the current size of the cache
	 * in term of object number currently managed by the cache.
	 */
	int getCurrentSize();

	/**
	 * @return an positive integer value. It is the current size of the cache
	 * in octets. This size matches to the memory foot print of the entries set
	 * managed by the cache. 
	 */
	int getCurrentMemorySize();

	/**
	 * @return an unmodifiable collection containing the identifiers of current
	 * cache entries.
	 */
	Collection getCurrentEntryIdentifiers();
}
