/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.cache.lib;

import org.objectweb.perseus.cache.replacement.api.ReplacementManager;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * This component is a thread managing the auto cleaning of a cache. The clean
 * up is launched when a threshold of the cache size is reached.
 *
 * @author S.Chassande-Barrioz
 */
public class BackgroundCleaner implements Runnable {

	private Object wakeup;
	private boolean runnning;
	private ReplacementManager rm;
	private Thread thread;
	private int size;
	private Logger logger;

	public BackgroundCleaner(ReplacementManager rm, Logger logger) {
		this.rm = rm;
		wakeup = new Object();
		runnning = true;
		this.logger = logger;
		thread = new Thread(this);
		thread.setDaemon(true);
		thread.start();
	}

	public Thread getThread() {
		return thread;
	}

	public void restart() {
		if (!runnning) {
			synchronized(wakeup) {
				if (!runnning) {
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
						logger.log(BasicLevel.INFO, "BackgroundCleaner restart");
					}
					runnning = true;
					wakeup.notify();
				}
			}
		}
	}

	public void stop() {
		if (runnning) {
			synchronized(wakeup) {
				if (runnning) {
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
						logger.log(BasicLevel.INFO, "BackgroundCleaner stop");
					}
					runnning = false;
					wakeup.notify();
				}
			}
		}
	}

	public void backgroungCleanup(int _size) {
		synchronized(wakeup) {
			this.size = _size;
			wakeup.notify();
		}
	}

	public void run() {
		while(runnning) {
			synchronized(wakeup) {
				try {
					if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG, "Wait next wake up");
					}
					wakeup.wait();
				} catch (InterruptedException e) {
				}
			}
			if (runnning) {
				if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
							"Cleanup the cache in background (" + size +")");
				}
				try {
					rm.forceFree(size);
				} catch (CacheException e) {
					if (logger != null) {
						logger.log(BasicLevel.ERROR,
						        "Impossible to cleanup the cache ("
						        + size + "): ", e);
					}
				}
			}
		}
		if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
			logger.log(BasicLevel.INFO, "BackgroundCleaner end");
		}
	}
}
