/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.cache.replacement.lib;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.perseus.cache.api.CacheEntryFilter;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.FixableCacheEntry;
import org.objectweb.perseus.cache.api.UnbindManager;
import org.objectweb.perseus.cache.api.CacheCapacityEventListener;
import org.objectweb.perseus.cache.api.CacheCapacityEvent;
import org.objectweb.perseus.cache.replacement.api.ReplaceableCacheEntry;
import org.objectweb.perseus.cache.replacement.api.ReplacementManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Collections;

/**
 * This class is a common implementation of a ReplacementManager.
 *
 * @author E.Bruneton
 * @see FIFOReplacementManager, LRUReplacementManager, MRUReplacementManager
 */
public abstract class AbstractReplacementManager
		implements ReplacementManager, BindingController, CacheCapacityEventListener {

	public final static String UNBIND_MANAGER_BINDING = "unbind-manager";

	protected UnbindManager ub;

	/**
	 * Sorted list of WeakReplaceableCacheEntry entries. Each entry must have a
	 * unique age, so that this sorted set behaves as a sorted list.
	 */
	protected SortedSet entries;
	protected HashMap oid2wentry;

	/**
	 * Queue used to register the weak references stored in the entries list.
	 */
	protected ReferenceQueue queue;

	protected Logger logger = null;

	public AbstractReplacementManager() {
		entries = new TreeSet();
		oid2wentry = new HashMap();
		queue = new ReferenceQueue();
	}

	//IMPLEMENTATION OF THE UserBindingControler INTERFACE //
	//------------------------------------------------------//

    public String[] listFc() {
        return new String[] {
            UNBIND_MANAGER_BINDING,
        };
    }

	public Object lookupFc(String s) {
		if (UNBIND_MANAGER_BINDING.equals(s)) {
			return ub;
		}
		return null;
	}

	public void bindFc(String s, Object o) {
		if (UNBIND_MANAGER_BINDING.equals(s)) {
			ub = (UnbindManager) o;
		} else if ("logger".equals(s)) {
			logger = (Logger) o;
		}
	}

	public void unbindFc(String s) {
		if (UNBIND_MANAGER_BINDING.equals(s)) {
			ub = null;
		}
	}

	//IMPLEMENTATION OF THE CacheCapacityEventListener INTERFACE //
	//---------------------------------------------------//

	public void cacheResized(CacheCapacityEvent event) {
		int size = event.getSize();
		if (size > 0) {
			synchronized(entries) {
				if (oid2wentry.size() == 0) {
					oid2wentry = new HashMap(size);
				}
			}
		}
	}

	//IMPLEMENTATION OF THE ReplacementManager INTERFACE //
	//---------------------------------------------------//

	public void addForReplacement(FixableCacheEntry entry)
			throws CacheException {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
				"add the entry " + entry.getCeIdentifier());
		ReplaceableCacheEntry dentry = check(entry);
		synchronized(entries) {
			clean();
			WeakReplaceableCacheEntry we = (WeakReplaceableCacheEntry)
				oid2wentry.get(entry.getCeIdentifier());
			add(dentry);
			if (we != null) {
				entries.remove(we);
				we.updateEntry((ReplaceableCacheEntry) entry);
			} else {
				we = new WeakReplaceableCacheEntry(dentry, queue);
				oid2wentry.put(entry.getCeIdentifier(), we);
			}
			entries.add(we);
		}
	}

	public void adjustForReplacement(FixableCacheEntry entry)
			throws CacheException {
		if (logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG,
				"touch the entry " + entry.getCeIdentifier());
		}
		ReplaceableCacheEntry dentry = check(entry);
		synchronized(entries) {
			clean();
			WeakReplaceableCacheEntry we = (WeakReplaceableCacheEntry)
				oid2wentry.get(entry.getCeIdentifier());
			touch(dentry);
			if (we != null) {
				entries.remove(we);
				we.updateEntry((ReplaceableCacheEntry) entry);
			} else {
				we = new WeakReplaceableCacheEntry(dentry, queue);
				oid2wentry.put(entry.getCeIdentifier(), we);
			}
			entries.add(we);
		}
	}

	public int forceFree(int capacity) throws CacheException {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "forceFree: requested: " + capacity);
		}
		// elects entries which could be evicted
		ArrayList oids = new ArrayList(capacity);
		int count = 0;
		synchronized(entries) {
			Iterator i = entries.iterator();
			while (count < capacity && i.hasNext()) {
				WeakReplaceableCacheEntry wentry = (WeakReplaceableCacheEntry) i.next();
				ReplaceableCacheEntry entry = (ReplaceableCacheEntry) wentry.get();
				if (entry != null && entry.getCeFixCount() == 0) {
					oids.add(entry.getCeIdentifier());
					count++;
				}
			}
		}
		if (debug) {
			logger.log(BasicLevel.DEBUG, "forceFree: try to unbound, oids=" + oids);
		}
		//Ask the eviction to the UnbindManager
		Collection _oids = ub.unbindAll(oids, false);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "forceFree: entries unbound, oid=" + _oids);
		}
		return _oids.size();
	}

	public void removeForReplacement(Object oid) {
		synchronized(entries) {
			Object we = oid2wentry.remove(oid);
			if (we != null) {
				entries.remove(we);
			}
		}
	}


	//IMPLEMENTATION OF THE UnbindManager INTERFACE //
	//----------------------------------------------//
	public Collection unbind(CacheEntryFilter filter, boolean force) throws CacheException {
	    ArrayList oids;
		synchronized(entries) {
			if (entries.isEmpty()) {
				return Collections.EMPTY_SET;
			}
			oids = new ArrayList();
			for(Iterator it = entries.iterator(); it.hasNext();) {
				WeakReplaceableCacheEntry wentry =
				        (WeakReplaceableCacheEntry) it.next();
				if (wentry != null) {
				    ReplaceableCacheEntry ce = (ReplaceableCacheEntry) wentry.get();
				    if (ce != null && filter.accept(ce)) {
				        oids.add(wentry.getCeIdentifier());
				    }
				}
			}
		}
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "unbindAll(oids, " + force
			        + ") try to unbound, oids=" + oids);
		}
		//Ask the eviction to the UnbindManager
		Collection res = ub.unbindAll(oids, force);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "entries unbound, oid=" + res);
		}
		return res;
	}
	public Collection unbindAll(Collection _oids, boolean force) throws CacheException {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		//Calculate the list of oid which match to unfixed entries
		ArrayList oids;
		synchronized(entries) {
			if (entries.isEmpty()) {
				return Collections.EMPTY_SET;
			}
			oids = new ArrayList();
			Iterator it = _oids.iterator();
			while(it.hasNext()) {
				Object oid = it.next();
				WeakReplaceableCacheEntry wentry =
				        (WeakReplaceableCacheEntry) oid2wentry.get(oid);
				if (wentry != null) {
					oids.add(oid);
				}
			}
		}
		if (debug) {
			logger.log(BasicLevel.DEBUG, "unbindAll(oids, " + force
			        + ") try to unbound, oids=" + oids);
		}
		//Ask the eviction to the UnbindManager
		_oids = ub.unbindAll(oids, force);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "entries unbound, oid=" + _oids);
		}
		return _oids;
	}

	public void unbindAll(boolean force) throws CacheException {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		synchronized(entries) {
			if (entries.isEmpty()) {
				return;
			}
			entries.clear();
			oid2wentry.clear();
		}
		if (debug) {
			logger.log(BasicLevel.DEBUG, "unbindAll(" + force
			        + ") try to unbound");
		}
		//Ask the eviction to the UnbindManager
		ub.unbindAll(force);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "all entries unbound");
		}
	}
	
	public Collection unbindUnfixed(boolean force) throws CacheException {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		//Calculate the list of oid matching to evictable entries
		ArrayList oids;
		synchronized(entries) {
			if (entries.isEmpty()) {
				return Collections.EMPTY_SET;
			}
			oids = new ArrayList();
			Iterator it = entries.iterator();
			while(it.hasNext()) {
				WeakReplaceableCacheEntry wentry = (WeakReplaceableCacheEntry) it.next();
				ReplaceableCacheEntry ce = (ReplaceableCacheEntry) wentry.get();
				if (ce == null || ce.getCeFixCount() == 0) {
					oids.add(wentry.getCeIdentifier());
				}
			}
		}
		if (debug) {
			logger.log(BasicLevel.DEBUG, "unbindAll(" + force
			        + ") try to unbound, oids=" + oids);
		}
		//Ask the eviction to the UnbindManager
		Collection _oids = ub.unbindAll(oids, force);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "entries unbound, oid=" + _oids);
		}
		return _oids;
	}

    /**
     * It forces the replacement manager to unbind a CacheEntry instance from
     * the cache.
     */
    public boolean unbind(Object oid, boolean force) throws CacheException {
	    boolean debug = logger.isLoggable(BasicLevel.DEBUG);
        if (oid == null) {
	        return false;
        }
	    if (debug)
            logger.log(BasicLevel.DEBUG, "unbind, oid=" + oid);
		WeakReplaceableCacheEntry we;
		try {
			//try out of synchronization
			we = (WeakReplaceableCacheEntry ) oid2wentry.get(oid);
		} catch (Throwable e) {
			// if a concurrency error is thrown, then try another with the
			// synchronization
			synchronized(entries) {
				we = (WeakReplaceableCacheEntry ) oid2wentry.get(oid);
			}
		}
        if (we != null) {
			return ub.unbind(oid, force);
        } else {
			logger.log(BasicLevel.WARN,
					"The entry has not been found from the cache");
	        return false;
		}
    }

	//---------------------------------------------------//

	private ReplaceableCacheEntry check(FixableCacheEntry entry) {
		try {
			return (ReplaceableCacheEntry) entry;
		} catch (ClassCastException e) {
			throw new IllegalArgumentException(
					"This replacement manager works only with ReplaceableCacheEntry entries");
		}
	}

	protected abstract void add(ReplaceableCacheEntry entry);

	protected abstract void touch(ReplaceableCacheEntry entry);

	/**
	 * Removes all entries that have been garbage collected. Does not block.
	 */
	private void clean() {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		while (true) {
			WeakReplaceableCacheEntry wentry =
				        (WeakReplaceableCacheEntry) queue.poll();
			if (wentry != null) {
				if (debug) {
					logger.log(BasicLevel.DEBUG, "auto cleanup, oid="
					        + wentry.getCeIdentifier());
				}
				entries.remove(wentry);
				oid2wentry.remove(wentry.getCeIdentifier());
			} else {
				return;
			}
		}
	}

	/**
	 * The entries stored in 'entries' are weak references to the real entries.
	 * Each weak entry also stores the age of the real entry at the time it was
	 * added in 'entries'.
	 */
	private static class WeakReplaceableCacheEntry
			extends WeakReference implements Comparable {
		private long date;
		private Object oid;

		public WeakReplaceableCacheEntry(ReplaceableCacheEntry ce, ReferenceQueue queue) {
			super(ce, queue);
			oid = ce.getCeIdentifier();
			date = ce.getCeAge();
		}

		public Object getCeIdentifier() {
			return oid;
		}

		public int compareTo(Object o) {
			long odate = ((WeakReplaceableCacheEntry) o).date;
			if (date == odate) {
				return 0;
			} else {
				return date < odate ? -1 : 1;
			}
		}

		public boolean equals(Object o) {
			return date == ((WeakReplaceableCacheEntry) o).date;
		}

		public void updateEntry(ReplaceableCacheEntry ce) {
			oid = ce.getCeIdentifier();
			date = ce.getCeAge();
		}
	}
}
